import io
import os
from typing import Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image as PILImage


def get_img_from_fig(fig, dpi=180, color_cvt_flag=cv2.COLOR_BGR2RGB) -> np.ndarray:
    """Make numpy array from mpl fig
    Parameters
    ----------
    fig : plt.Figure
        Matplotlib figure, usually the result of plt.imshow()
    dpi : int, optional
        Dots per inches of the image to save. Note, that default matplotlib
        figsize is given in inches. Example: px, py = w * dpi, h * dpi  pixels
        6.4 inches * 100 dpi = 640 pixels, by default 180
    color_cvt_flag : int, optional
        OpenCV cvtColor flag. to get grayscale image,
        use `cv2.COLOR_BGR2GRAY`, by default `cv2.COLOR_BGR2RGB`.
    Returns
    -------
    np.ndarray[np.uint8]
        Image array
    """

    with io.BytesIO() as buffer:
        fig.savefig(buffer, format="png", dpi=dpi)
        buffer.seek(0)
        img_arr = np.frombuffer(buffer.getvalue(), dtype=np.uint8)
    return cv2.cvtColor(cv2.imdecode(img_arr, 1), color_cvt_flag)


def array_to_uintimage(array: np.ndarray, cvt_flag=cv2.COLOR_BGR2RGB) -> np.ndarray:
    """Converts any array to uint8 data type format using a magic of the matplotlib.
    Parameters
    ----------
    array : np.ndarray
        Image array.
    cvt_flag : int, optional
        OpenCV cvtColor flag. to get grayscale image,
        use `cv2.COLOR_BGR2GRAY`, by default `cv2.COLOR_BGR2RGB`
    Returns
    -------
    np.ndarray[np.uint8]
        Output image array in uint8 datatype
    """

    w, h = array.shape
    dpi = 90
    fig = plt.figure(figsize=(h / dpi, w / dpi))
    plt.imshow(array, cmap="gray")
    plt.axis("off")
    plt.tight_layout(pad=0)
    uint_img = get_img_from_fig(fig, dpi=dpi, color_cvt_flag=cvt_flag)
    plt.close(fig)
    return uint_img


def open_with_PIL(
    image: Union[os.PathLike, PILImage.Image, np.ndarray]
) -> PILImage.Image:
    """Reads image to PIL.Image.
    Parameters
    ----------
    image : Union[os.PathLike, PIL.Image.Image, np.ndarray]
        Image path, PIL.Image or numpy array.
    Returns
    -------
    PIL.Image.Image
        Output image as PIL.Image object
    Raises
    ------
    AttributeError
        In case of inappropriate input.
    """

    if isinstance(image, (str, os.PathLike)):
        pil_image = PILImage.open(image).convert("RGB")
    elif isinstance(image, PILImage.Image):
        pil_image = image.convert("RGB")
    elif isinstance(image, np.ndarray):
        image = image.squeeze()
        assert image.ndim == 2, "Supported only grayscale numpy arrays"
        image_array = array_to_uintimage(image)
        pil_image = PILImage.fromarray(image_array).convert("RGB")
        pil_image = pil_image.resize(image.shape[::-1])
    else:
        raise AttributeError("Unsupported type of image input")
    return pil_image



from typing import Tuple, Union, List

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as colormap


RGBAColor = Union[
    Tuple[np.uint8, np.uint8, np.uint8, np.uint8], Tuple[float, float, float, float]
]


class LabelColor:
    def __init__(
        self,
        num_labels: int,
        mpl_colormap: str = "gist_rainbow",
        alpha: float = 0.3,
        mode: str = "uint8",
        return_legend_color: bool = False,
    ):
        """Colormap class. Easy-peazy way to get colos for your segmentation labels.
        Parameters
        ----------
        num_labels : int
            Number of labels to map with colors.
        mpl_colormap : str, optional
            Matplotlib colormap name.
            See all at: https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
            , by default "gist_rainbow"
        alpha : float, optional
            Opacity, by default 0.3
        mode : str, optional
            Type of image values. Either uint8 or float, by default "uint8"
        return_legend_color: bool, optional
            If true, getitem retirns tuple of mask color
            and same less opaque legend color.
        """

        self.cmap = colormap.get_cmap(mpl_colormap, num_labels)
        self.alpha = alpha
        self.mode = mode
        self.num_labels = num_labels
        self.return_legend_color = return_legend_color

    def __getitem__(self, i: int) -> Union[RGBAColor, Tuple[RGBAColor, RGBAColor]]:
        color: RGBAColor
        if (i < 0) or (i >= self.num_labels):
            raise IndexError(
                "Label index out of scope, set larger `num_labels`."
                + f" Current: {self.num_labels}"
            )

        if self.mode == "float":
            color = self.cmap(i, self.alpha)
        elif self.mode == "uint8":
            color = np.array(self.cmap(i, self.alpha))
            color = tuple(np.array(color * 255, dtype=np.uint8))
        else:
            raise AttributeError("Unsupported mode. Use 'uint8' or 'float'")

        if self.return_legend_color:
            return color, self.get_legend_color(i)
        else:
            return color

    def __len__(self):
        return self.num_labels

    def __iter__(self):
        return LabelColorIter(self)

    def get_legend_color(self, i, alpha=0.8):
        """Return more opaque version of the same color
        Parameters
        ----------
        i : int
            color index
        alpha : float, optional
            Opaque coefficient, by default 0.8
        Returns
        -------
        tuple
            RGBA color in tuple.
        """

        return self.cmap(i, alpha)

    @staticmethod
    def blend(*args: RGBAColor) -> RGBAColor:
        """Blends colors.
        Returns
        -------
        RGBAColor
        """

        if isinstance(args[0][0], int):
            return tuple(np.array(args, dtype="uint16").mean(0).astype("uint8"))
        else:
            return tuple(np.array(args).mean(0))


class LabelColorIter:
    def __init__(self, label_color: LabelColor) -> None:
        self.label_color = label_color
        self._i = 0

    def __next__(self) -> Union[RGBAColor, Tuple[RGBAColor, RGBAColor]]:
        if self._i < len(self.label_color):
            color: Union[RGBAColor, Tuple[RGBAColor, RGBAColor]] = self.label_color[
                self._i
            ]
        else:
            raise StopIteration
        self._i += 1
        return color

    def __iter__(self):
        return self



import os
from typing import Optional, Tuple, Union, List

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from PIL import Image as PILImage

from .label_color import LabelColor
from .utils import open_with_PIL


def overlay_masks(
    image: Union[os.PathLike, PILImage.Image, np.ndarray],
    boolean_masks: List[np.ndarray],
    labels: Optional[List[str]] = None,
    colors: np.ndarray = None,
    figsize: Optional[Tuple[int, int]] = None,
    dpi: int = 90,
    mask_alpha: float = 0.4,
    mpl_colormap: str = "tab20",
) -> plt.Figure:
    """Overlays masks on the image.
    Parameters
    ----------
    image : Union[str, PIL.Image.Image, np.ndarray]
        Image path or PIl.Image or numpy array. If image size inconsistent with
        the masks size, image will be resized.
    boolean_masks : List[np.ndarray[bool]]
        List of segmentation masks. All masks should be the same size, equal to size of the image.
    labels : Optional[List[str]], optional
        Optional label names. Provide in the same order as the corresponding masks.
        If not provided, will be set as range(len(boolean_masks)), by default None
    colors : np.ndarray
        Array of shape (n_labels x 4). Example to get persistent colormap: `plt.cm.tab20(np.arange(NUM_LABELS))`
    figsize : tuple, optional
        Size in inches of the output image, by default (12, 12)
    dpi : int, optional
        Resolution of the output image. Note: 'px, py = w * dpi, h * dpi', by default 120
    mask_alpha : float, optional
        Masks opaque value, by default 0.4
    Returns
    -------
    plt.Figure
        Output figure with masks legend.
    """

    if labels is not None:
        assert len(labels) == len(
            boolean_masks
        ), "Number of provided labels != number of masks"
    else:
        labels = [f"mask_{_}" for _ in range(len(boolean_masks))]

    if colors is not None:
        assert len(colors) == len(
            boolean_masks
        ), "Number of provided colors != number of masks"

    image = open_with_PIL(image)
    image_size = tuple(np.array(image.size)[::-1])

    assert all(
        mask.shape == image_size for mask in boolean_masks
    ), "Label mask size is not equal to image size"

    if colors is None:
        cbar = LabelColor(
            num_labels=len(boolean_masks),
            alpha=mask_alpha,
            return_legend_color=True,
            mpl_colormap=mpl_colormap,
        )

    else:
        mask_colors = colors.copy()
        mask_colors[:, -1] *= mask_alpha
        mask_colors = (mask_colors * 255).astype("uint8")
        cbar = zip(mask_colors, colors)

    segmentation_overlay = np.zeros((*image_size, 4), dtype=np.uint16)
    segmentation_mask = np.zeros(image_size, dtype=bool)
    legend_elements = []

    for mask, label, (color, legend_color) in zip(boolean_masks, labels, cbar):

        assert mask.dtype == "bool"

        intersection = mask & segmentation_mask
        segmentation_mask = mask | segmentation_mask

        # Paint non-overlapping area
        segmentation_overlay[mask ^ intersection] = color

        # Blend overlapping area
        segmentation_overlay[intersection] = (
            segmentation_overlay[intersection] + color
        ) / 2

        legend_elements.append(Patch(color=legend_color, label=label))

    segmentation_overlay = PILImage.fromarray(segmentation_overlay.astype("uint8"))
    image.paste(segmentation_overlay, mask=segmentation_overlay)

    fig = plt.figure(figsize=figsize, dpi=dpi)
    plt.imshow(image)
    plt.axis("off")
    plt.legend(
        handles=legend_elements,
        loc="upper left",
        frameon=False,
        bbox_to_anchor=(1.01, 1),
    )
    plt.subplots_adjust(left=0.8)
    plt.tight_layout()

    return fig



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="segmentation-mask-overlay",
    version="0.1.0",
    description="Plotting the segmentation masks has never been so exciting!",
    long_description=README,
    long_description_content_type="text/markdown",
    author="armavox",
    author_email="armavox@gmail.com",
    license="MIT",
    url="https://github.com/lobantseff/segmentation-mask-overlay",
    packages=find_packages(include=["segmentation_mask_overlay"]),
    include_package_data=True,
    install_requires=[
        "matplotlib>=3.4.2",
        "numpy>=1.20",
        "opencv-python>=4.5.3",
        "Pillow>=7.2.0"

    ],
    setup_requires=["flake8"],
)
